/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * Copyright (c) 2003-2005 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/FSCL.txt
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author(s): Josh Wilsdon <josh@wizard.ca>
 *
 * CVS Id: $Header: /web/programmers/cvs/magicmail-backend/builddb/builddb.c,v 1.7 2005/12/19 21:13:49 josh Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */

#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <db.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include "builddb.h"
#include "magicdb.h"

DB *dbp;

char segfaultdir[] = "/var/cores/build_dbfiles/XXXXX";

void msd_sigsegfault(int arg)
{                
    arg=arg;

    if ((mkdir(segfaultdir, 0700) == 0) && (chdir(segfaultdir) == 0)) {
        /* abort so we can get a core dump */
        abort();
    } else {
        fprintf(stderr, "segfault: failed to dump core\n");
    }

    exit(1);
}

int main()
{
    void *handle;
    char *error;
    int (*init) (void);
    int (*cleanup) (void);
    int (*foreachRecord) (int type,
                          void (*func) (int type, int keysize, char *key,
                                        int datasize, char *data));
    int segfaultdirlen;
    struct rlimit rlim;

    /* setup for segfault core dump */
    segfaultdirlen = strlen(segfaultdir);
    if (segfaultdirlen > 5) {
        snprintf(segfaultdir + (segfaultdirlen - 5), 6, "%05d", (int)getpid());
    }
    rlim.rlim_cur = RLIM_INFINITY;
    rlim.rlim_max = RLIM_INFINITY;
    setrlimit(RLIMIT_CORE, &rlim);

    /* setup a segfault handler so we might get a message printed if we crash */
    signal(SIGSEGV, msd_sigsegfault);

    /* first try /usr/lib/mod_postgresql.so */
    handle = dlopen("/usr/lib/mod_postgresql.so", RTLD_LAZY);
    if (handle == NULL) {
        /* try again with /usr/local/modules */
        handle = dlopen("/usr/local/modules/mod_postgresql.so", RTLD_LAZY);
        if (handle == NULL) {
            fprintf(stderr, "dlopen(): %s\n", dlerror());
            exit(1);
        }
    }

    init = dlsym(handle, "init");
    if ((error = dlerror()) != NULL) {
        fprintf(stderr, "dlsym(): %s\n", error);
        exit(1);
    }
    init();

    foreachRecord = dlsym(handle, "foreachRecord");
    if ((error = dlerror()) != NULL) {
        fprintf(stderr, "dlsym(): %s\n", error);
        exit(1);
    }

    if (openDB(SPAM_RULES) != 0) {
        fprintf(stderr, "db_open failed\n");
        exit(1);
    }
    if (foreachRecord(SPAM_RULES, writeRecord) != 0) {
        fprintf(stderr, "foreachRecord(SPAM) failed\n");
        exit(1);
    }
    // printf("countPairs(%s): %d\n", SPAM_DATABASE, countPairs());
    closeDB();

    if (openDB(USER_INFO) != 0) {
        fprintf(stderr, "db_open failed\n");
        exit(1);
    }
    if (foreachRecord(USER_INFO, writeRecord) != 0) {
        fprintf(stderr, "foreachRecord(USER) failed\n");
        exit(1);
    }
    // printf("countPairs(%s): %d\n", USER_DATABASE, countPairs());
    closeDB();

    cleanup = dlsym(handle, "cleanup");
    if ((error = dlerror()) != NULL) {
        fprintf(stderr, "dlsym(): %s\n", error);
        exit(1);
    }
    cleanup();

    dlclose(handle);

    exit(0);
}

/* 
 * According to the documentation, DB->close() can fail to sync or on other
 * catastrophic conditions.  If it fails, we cannot be guaranteed the state
 * of the database, so we exit(1) to fail the whole run.
 *
 */
void closeDB(void)
{
    if (dbp->close(dbp, 0) != 0) {
        fprintf(stderr, "db_close failed\n");
        exit(1);
    }
}

#if (DB_VERSION >= 3)
int countPairs(void)
{
    DB_HASH_STAT *stats;

    if (dbp->stat(dbp, &stats, NULL, 0) == 0) {
        return ((int) stats->hash_ndata);
    }

    return (-1);
}
#endif

int verifyDB(const char *dbname)
{
    if (dbname == NULL) {
        return (1);
    }

#if (DB_VERSION_MAJOR >= 3) 
    return (dbp->verify(dbp, dbname, NULL, NULL, 0));
#else
    #warning "integrity cannot be checked!"
    return (0);
#endif
}

int createDB(void)
{
#if (DB_VERSION_MAJOR >= 3)
    int ret;

    if ((ret = db_create(&dbp, NULL, 0)) != 0) {
        fprintf(stderr, "db_create: %s\n", db_strerror(ret));
        return (1);
    }
#endif

    return (0);
}

int openDB(int type)
{
    char *filename;
    int ret;

    switch (type) {
        case SPAM_RULES:
            filename = SPAM_DATABASE;
            break;
        case USER_INFO:
            filename = USER_DATABASE;
            break;
        default:
            fprintf(stderr, "openDB: unknown type: %d\n", type);
            return (1);
            break;
    }

    /* Create the database handle. */
    if (createDB() == 0) {

#if (DB_VERSION_MAJOR < 2 || DB_VERSION_MAJOR > 4)
	#error "This release only supports Berkeley DB version 2-4"
#elif (DB_VERSION_MAJOR == 2)
	if ((ret = db_open(filename, DB_HASH, DB_CREATE, 0664, NULL, NULL, &dbp)) == 0) {
#elif (DB_VERSION_MAJOR == 3)
        /* Open the underlying database. */
        if ((ret = dbp->open(dbp, filename, NULL, DB_HASH, DB_CREATE, 0664)) == 0) {
#elif (DB_VERSION_MAJOR == 4)
        if ((ret = dbp->open(dbp, NULL, filename, NULL, DB_HASH, DB_CREATE, 0664)) == 0) {
#endif
            return (0);
        } else {
#if (DB_VERSION_MAJOR >= 3) 
            dbp->err(dbp, ret, "%s", filename);
#else
	    fprintf(stderr, "dbp->open(): error %d\n", ret);
#endif
        }
    }

    return (1);
}

void writeRecord(int type, int keysize, char *key, int datasize, char *data)
{
    DBT dbkey, dbdata;
    int ret;

    type = type;                /* remove gcc warning */

    memset(&dbkey, 0, sizeof(dbkey));
    memset(&dbdata, 0, sizeof(dbdata));
    dbkey.data = key;
    dbkey.size = keysize;
    dbdata.data = data;
    dbdata.size = datasize;

    /* Store a key/data pair. */
    if ((ret = dbp->put(dbp, NULL, &dbkey, &dbdata, 0)) != 0) {
#if (DB_VERSION_MAJOR >= 3)
        dbp->err(dbp, ret, "DB->put");
#else
	fprintf(stderr, "dbp->put(): error %d\n", ret);
#endif
        exit(1);
    }

    return;
}
