/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * Copyright (c) 2002-2005 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author(s): Burton Samograd <burton@wizard.ca>
 *            Josh Wilsdon <josh@wizard.ca>
 *
 * CVS Id: $Header: /web/programmers/cvs/magicmail-backend/magic-smtpd/smtp_auth.c,v 1.10 2006/01/10 00:15:06 josh Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */

/* C includes */
#include <string.h>

/* Project includes */
#include "smtp.h"
#include "smtp_auth.h"
#include "magic-smtpd.h"
#include <liblm.h>
#include <common/base64.h>
#include <common/controls.h>
#include <common/load_controls.h>

int lmprintf(const char *format, ...);
int msd_timeout_read_line_lm(int fd, lm_string_t * line);

extern config_t config;
extern config_type_t config_type_info[];


static int msd_do_auth_helper(int type, lm_string_t username, lm_string_t password);


int msd_do_auth(int type, lm_string_t username64)
{
    char * username = NULL;
    char * password = NULL;
    lm_string_t lmusername = NULL;
    lm_string_t lmpassword = NULL;
    lm_string_t password64 = NULL;
    int retval = 0;

    /*
     * TODO: when we do this properly (modules) we'll need to have a function 
     *       supported_types() which will tell us which are supported.  We can
     *       then check to see if this is one of them.
     *
     */

    /* Only support LOGIN auth type for now */
    if (type != SMTP_AUTH_LOGIN) {
        syslog(LOG_INFO, "Unsupported authorization type requested (not LOGIN)");
        return LM_AUTH_UNSUPPORTED;
    }

    if (username64 == NULL) {
        /* Print the Username: prompt in base64, read an input line and decode 
         * it to get the username */
        lmprintf("334 VXNlcm5hbWU6\r\n"); /* Username: */
        msd_timeout_read_line_lm(0, &username64);
        LM_STRING_CHOP(username64);
        /* check for user cancel '*' */
        if (LM_STRING_LEN(username64) == 1
            && LM_STRING_BUFFER(username64)[0] == '*') {
            /* user canceled */
            LM_STRING_FREE(username64);
            return LM_AUTH_CANCEL;
        }
    }
    if (b64decode(LM_STRING_BUFFER(username64), &username)) {
        LM_STRING_FREE(username64);
        return LM_AUTH_BAD_DATA;
    } else {
        LM_STRING_NEW(lmusername, username);
        LM_STRING_CHOP(lmusername);
    }
    LM_STRING_FREE(username64);

    /* Print the Password: prompt in base64, read an input line and decode it to
     * get the password */
    lmprintf("334 UGFzc3dvcmQ6\r\n"); /* Password: */
    msd_timeout_read_line_lm(0, &password64);
    LM_STRING_CHOP(password64);
    /* check for user cancel '*' */
    if (LM_STRING_LEN(password64) == 1
        && LM_STRING_BUFFER(password64)[0] == '*') {
        /* user canceled */
        free(username);
        LM_STRING_FREE(lmusername);
        LM_STRING_FREE(password64);
        return LM_AUTH_CANCEL;
    }
    if (b64decode(LM_STRING_BUFFER(password64), &password)) {
        /* failure to decode */
        free(username);
        LM_STRING_FREE(lmusername);
        LM_STRING_FREE(password64);
        return LM_AUTH_BAD_DATA;
    } else {
        LM_STRING_NEW(lmpassword, password);
        LM_STRING_CHOP(lmpassword);
    }
    LM_STRING_FREE(password64);

    retval = msd_do_auth_helper(type, lmusername, lmpassword);

    if (retval == LM_AUTH_OK) {
        /* set the relayclient variable and set remoteinfo to the username */
        if (config.tcprelayclient == NULL) {
            config.tcprelayclient = strdup("");
            if (config.tcprelayclient == NULL) {
                syslog(LOG_ERR, "msd_do_auth(): out of memory setting relayclient");
            }
        }

        if (config.tcpremoteinfo) {
            config.tcpremoteinfo = strdup(LM_STRING_BUFFER(lmusername));
            if (config.tcpremoteinfo == NULL) {
                syslog(LOG_ERR, "msd_do_auth(): out of memory setting tcpremoteinfo");
            }
        }

        syslog(LOG_INFO, "AUTH success: [%s]", LM_STRING_BUFFER(lmusername));
    } else {
        syslog(LOG_INFO, "AUTH failure: [%s]", LM_STRING_BUFFER(lmusername));
    }

    free(username);
    LM_STRING_FREE(lmusername);   /* don't need this anymore */
    free(password);
    LM_STRING_FREE(lmpassword);   /* don't need this anymore */

    return (retval);
}

int msd_do_auth_helper(int type, lm_string_t username, lm_string_t password)
{
    int ret;

#ifdef USE_LM_DATABASE
    if (config.use_database != 0) {
        syslog(LOG_DEBUG, "trying AUTH against database");
        ret = msd_do_auth_db(type, username, password);

        switch (ret) {
            case LM_AUTH_OK:
                /* FALLTHROUGH */
            case LM_AUTH_FAILED:
                return (ret);
                break;
            case LM_AUTH_NO_USER:
                /* FALLTHROUGH */
            case LM_AUTH_ERROR:
                /* FALLTHROUGH */
            case LM_AUTH_ERROR_DB:
                /* FALLTHROUGH */
            default:
                /* pass through to next method */
                break;
        }
    }
#endif

    /* external program method is always available */
    if (config.ext_check_passwd_prog != NULL) {
        syslog(LOG_DEBUG, "trying AUTH against external program");
        ret = msd_do_auth_chkp(type, LM_STRING_BUFFER(username), LM_STRING_BUFFER(password));

        switch (ret) {
            case LM_AUTH_OK:
                /* FALLTHROUGH */
            case LM_AUTH_FAILED:
                return (ret);
                break;
            case LM_AUTH_BAD_DATA:
                /* FALLTHROUGH */
            case LM_AUTH_ERROR:
                /* FALLTHROUGH */
            case LM_AUTH_UNSUPPORTED:
                /* FALLTHROUGH */
            default:
                /* pass through to next method */
                break;
        }
    }

#ifdef USE_LM_DBFILE
    if (config.use_dbfile != 0) {
        syslog(LOG_DEBUG, "trying AUTH against dbfile");
        ret = msd_do_auth_dbfile(type, username, password);

        switch (ret) {
            case LM_AUTH_OK:
                /* FALLTHROUGH */
            case LM_AUTH_FAILED:
                return (ret);
                break;
            case LM_AUTH_NO_USER:
                /* FALLTHROUGH */
            case LM_AUTH_ERROR:
                /* FALLTHROUGH */
            default:
                /* pass through to next method */
                break;
        }
    }
#endif

    return (LM_AUTH_ERROR);
}
