/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * Copyright (c) 2002-2005 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/FSCL.txt
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author(s): Josh Wilsdon <josh@wizard.ca>
 *
 * CVS $Header: /web/programmers/cvs/magicmail-backend/magic-smtpd/tls.c,v 1.6 2005/12/22 00:50:41 josh Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */


#include <fcntl.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include <common/controls.h>
#include "tls.h"

extern int using_tls;
extern SSL *ssl;
extern config_t config;
extern char *encryption;

/* prototypes of functions in magic-smtpd.c */
int lmprintf(const char *format, ...);
int msd_quit_callback(int retval);
int msd_rset_callback(void);


#if (OPENSSL_VERSION_NUMBER < 0x00906000L)
#error OPENSSL Version 0.9.6 or higher required
#endif


#if 0
/* temporarily disabled */
RSA *tmp_rsa_cb(SSL *ssl, int export, int keylen)
{
    FILE *in;

    // XXX do we care about export controls?

    in = fopen(config.tls_rsakey, "r");

    if (in) {
        RSA *rsa = PEM_read_RSAPrivateKey(in, NULL, NULL, NULL);
        fclose(in);
        if (rsa) return rsa;
    }

    syslog(LOG_WARNING, "WARNING generating %d bit RSA key, this will be slow", keylen);

    return (RSA_generate_key(keylen, RSA_F4, NULL, NULL));
}
#endif

DH *tmp_dh_cb(SSL *ssl, int export, int keylen)
{
    FILE *in;

    ssl = ssl; /* silence GCC warning */
    export = export; /* silence GCC warning */

    // XXX do we care about export controls?

    in = fopen(config.tls_dhparams, "r");
    if (in) {
        DH *dh = PEM_read_DHparams(in, NULL, NULL, NULL);
        fclose(in);
        if (dh) return dh;
    }

    syslog(LOG_WARNING, "WARNING generating %d bit DH key, this will be slow", keylen);

    return (DH_generate_parameters(keylen, DH_GENERATOR_2, NULL, NULL));
}

int password_cb(char *buf, int num, int rwflag, void *userdata) 
{
    rwflag = rwflag;    /* silence GCC warning */
    userdata = userdata;    /* silence GCC warning */

    if (config.tls_password != NULL)  {
        /* password not defined */
        return (0);
    }

    if (num < (int)(strlen(config.tls_password) + 1)) {
        /* buffer is not big enough */
        return (0);
    }

    strcpy(buf, config.tls_password);

    return (strlen(config.tls_password));
}

int initialize_ssl()
{
    SSL_CTX *ctx;
    int timeout;

    SSL_library_init();
    SSL_load_error_strings();
    ctx=SSL_CTX_new(SSLv23_method());
    SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);

    /* Load the CAs */
    if(!(SSL_CTX_load_verify_locations(ctx, config.tls_cafile, config.tls_cadir))) {
        syslog(LOG_ERR, "OpenSSL(SSL_CTX_load_verify_locations): [%s]\n", ERR_error_string(ERR_get_error(), NULL));
    } 

    /* Load our certs and keys */
    if(!(SSL_CTX_use_certificate_chain_file(ctx, config.tls_certificate))) {
        syslog(LOG_ERR, "OpenSSL(SSL_CTX_use_certificate_chain_file): [%s]\n", ERR_error_string(ERR_get_error(), NULL));
    }

    if (config.tls_password != NULL) {
        SSL_CTX_set_default_passwd_cb(ctx, password_cb);
    }

    ssl = SSL_new(ctx);
        
    SSL_set_client_CA_list(ssl, NULL);
    SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);
    SSL_set_accept_state(ssl);

    /* 
     * Disabled until we can allow the user to cause this to be used
     *
     * SSL_set_tmp_rsa_callback(ssl, tmp_rsa_cb);
     */
    SSL_set_tmp_dh_callback(ssl, tmp_dh_cb);
    SSL_set_rfd(ssl, 0);    /* read from stdin */
    SSL_set_wfd(ssl, 1);    /* write to stdout */

    if (!SSL_use_RSAPrivateKey_file(ssl, config.tls_keyfile, SSL_FILETYPE_PEM)) {
        syslog(LOG_ERR, "OpenSSL(SSL_use_RSAPrivateKey_file): [%s]\n", ERR_error_string(ERR_get_error(), NULL));
    }

    /* turn off blocking for stdin and stdout */
    if (fcntl(0,F_SETFL,fcntl(0,F_GETFL,0) | O_NONBLOCK) == -1) {
        return (-1);
    }
    if (fcntl(1,F_SETFL,fcntl(1,F_GETFL,0) | O_NONBLOCK) == -1) {
        return (-1);
    }

    if (ssl_timeoutio(SSL_accept, NULL, 0, &timeout) <= 0) {
        if (timeout == 1) {
            /* timeout */
            syslog(LOG_ERR, "OpenSSL(SSL_accept) timed out");
            return (-1);
        } else {
            /* ERROR */
            syslog(LOG_ERR, "OpenSSL(SSL_accept): [%s]", ERR_error_string(ERR_get_error(), NULL));
            return (-1);
        }
    }

    /* turn off nonblock on stdout and stderr */
    fcntl(0,F_SETFL,fcntl(0,F_GETFL,0) & ~O_NONBLOCK);
    fcntl(1,F_SETFL,fcntl(1,F_GETFL,0) & ~O_NONBLOCK);

    return (0);
}


int ssl_timeoutio(int (*fun)(), char *buf, int len, int *timeout)
{
    int ret, changed = 0;
    time_t end, remain; 
    fd_set fds;
    struct timeval tv;
    
    end = time(NULL) + config.timeoutsmtpd;

    *timeout = 0;
    do {
        if (buf == NULL) {
            ret = fun(ssl);
        } else {
            ret = fun(ssl, buf, len);
        }
        if (ret > 0) {
            return (ret);
        }
  
        remain = end - time(NULL);
        if (remain < 0) {
            /* exit do{} while; */
            break;
        }
    
        tv.tv_sec = remain; 
        tv.tv_usec = 0;
    
        FD_ZERO(&fds);
        switch (SSL_get_error(ssl, ret)) {
            default: 
                return ret; /* some other error */
            case SSL_ERROR_WANT_READ:
                FD_SET(0 /* stdin */, &fds); 
                changed = select(0 /* stdin */ + 1, &fds, NULL, NULL, &tv);
                break;
            case SSL_ERROR_WANT_WRITE:
                FD_SET(1 /* stdout */, &fds); 
                changed = select(1 /* stdout */ + 1, NULL, &fds, NULL, &tv);
                break;
        }
    } while (changed > 0);
   
    if (changed != -1) {
        /* timeout */
        *timeout = 1;
    }

    return -1;
}

